; 
    include \masm32\include\masm32rt.inc
    .686
    .xmm
    include timers.asm

; 

; Thanks to: 
;  Mark Larson (for helping me optimizing the macros)

; This program uses the ranrot functions of Agner Fog (modified for the purpose of benchmarking)
;
; Original license text:
;  This random number generator software is copyrighted  2002-2005 by Agner Fog.
;  The software may be copied and used freely under the conditions of the GNU General
;  Public License with the following additional ethical constraint:
;
; This software shall not be used for any gambling application that is suitable
; for the exploitation of persons with ludomania for economic profit.
;
; Commercial licences are available on request. 

; WRandomInit must be called before the first call to any of the random
; functions.
; WRandom returns a floating point number between 0 and 1.
; WBRandom returns an integer between 0 and 0xFFFFFFFF
; WIRandom returns an integer in the interval

; Generates a 64-bit uniform random number
; Uses registers EAX-EDX
; Result in EDX:EAX
M_WBRandom MACRO

	MOV EBX, [p1]       ; ring buffer pointers
	MOV ECX, [p2]       ; ring buffer pointer
	MOV EDX, [randbuf][EBX]
	MOV EAX, [randbuf][EBX+4]       
	ROL EDX, R1                  ; rotate bits
	ROL EAX, R2
	ADD EDX, [randbuf][ECX]      ; add two dwords
	ADD EAX, [randbuf][ECX+4]
	MOV [randbuf][EBX], EAX      ; save in swapped order
	MOV [randbuf][EBX+4], EDX
	SUB EBX, 8                   ; decrement p1
	JNC @F
	MOV EBX, (KK-1)*8            ; wrap around p1
@@:
	SUB ECX, 8                   ; decrement p2
	JNC @F
	MOV ECX, (KK-1)*8            ; wrap around p2
@@:
	MOV [p1], EBX                ; save updated pointers
	MOV [p2], ECX

ENDM

; This version uses a lookup table to adjust it's pointers
M_WBRandomL MACRO	; Uses lookup table

	MOV EBX, [p1]       ; ring buffer pointers
	MOV ECX, [p2]       ; ring buffer pointer
	MOV EDX, [randbuf][EBX]
	MOV EAX, [randbuf][EBX+4]       
	ROL EDX, R1                  ; rotate bits
	ROL EAX, R2
	ADD EDX, [randbuf][ECX]      ; add two dwords
	ADD EAX, [randbuf][ECX+4]
	MOV [randbuf][EBX], EAX      ; save in swapped order
	MOV [randbuf][EBX+4], EDX
	mov ebx,[PtrNext+ebx*4]
	mov ecx,[PtrNext+ecx*4]
	MOV [p1], EBX                ; save updated pointers
	MOV [p2], ECX

ENDM

; This version generates only one doubleword and reuses the second
; Useful if you are only interested in 32-bit numbers
M_WBRandom2 MACRO

local fred
	mov eax,[last_value]
	test eax,eax
	je @F
	xor edx,edx
	mov [last_value],edx
	jmp fred
	
@@:
	MOV EBX, [p1]       ; ring buffer pointers
	MOV ECX, [p2]       ; ring buffer pointer
	MOV EDX, [randbuf][EBX]
	MOV EAX, [randbuf][EBX+4]       
	ROL EDX, R1                  ; rotate bits
	ROL EAX, R2
	ADD EDX, [randbuf][ECX]      ; add two dwords
	ADD EAX, [randbuf][ECX+4]
	MOV [randbuf][EBX], EAX      ; save in swapped order
	MOV [randbuf][EBX+4], EDX
	SUB EBX, 8                   ; decrement p1
	JNC @F
	MOV EBX, (KK-1)*8            ; wrap around p1
@@:
	SUB ECX, 8                   ; decrement p2
	JNC @F
	MOV ECX, (KK-1)*8            ; wrap around p2
@@:
	MOV [p1], EBX                ; save updated pointers
	MOV [p2], ECX

	mov [last_value],edx
fred:
ENDM

; This version generates only one doubleword and reuses the second
; Useful if you are only interested in 32-bit numbers
; Uses the lookup table
M_WBRandom2L MACRO
	
local fred
	mov eax,[last_value]
	test eax,eax
	je @F
	xor edx,edx
	mov [last_value],edx
	jmp fred
	
@@:
	MOV EBX, [p1]       ; ring buffer pointers
	MOV ECX, [p2]       ; ring buffer pointer
	MOV EDX, [randbuf][EBX]
	MOV EAX, [randbuf][EBX+4]       
	ROL EDX, R1                  ; rotate bits
	ROL EAX, R2
	ADD EDX, [randbuf][ECX]      ; add two dwords
	ADD EAX, [randbuf][ECX+4]
	MOV [randbuf][EBX], EAX      ; save in swapped order
	MOV [randbuf][EBX+4], EDX
	mov ebx,[PtrNext+ebx*4]
	mov ecx,[PtrNext+ecx*4]
	MOV [p1], EBX                ; save updated pointers
	MOV [p2], ECX

	mov [last_value],edx
fred:
ENDM

; Generates a number between 0.0 and 1.0
M_WRandom MACRO

	M_WBRandom
	
	OR	EDX, 80000000H           ; fast convert to float
	MOV DWORD PTR [randp1],EAX
	MOV DWORD PTR [randp1+4],EDX
	FLD1
	FLD	[randp1]
	FSUBR

ENDM

M_WIRandom MACRO min:req, max:req

	M_WBRandom
	
	MOV EBX, max - min + 1
	MOV ECX, EDX             ; high bits of random number
	MUL EBX                  ; multiply low 32 bits
	MOV EAX, ECX
	MOV ECX, EDX
	MUL EBX                  ; multiply high 32 bits
	ADD EAX, ECX
	ADC EDX, min       ; add min
	MOV EAX, EDX

ENDM

M_WDRandom MACRO

	M_WBRandom
	
	and edx, 000FFFFFh
	or edx, 3FF00000h
	
	mov dword ptr tmpdouble, eax
	mov dword ptr tmpdouble + 4, edx
	
	movlpd xmm0, tmpdouble
	subsd xmm0, double_one
	movlpd tmpdouble, xmm0

ENDM

M_WSRandom MACRO

	M_WBRandom2L
	
	and eax, 007FFFFFh
	or eax, 3F800000h
	movd xmm0, eax
	subss xmm0, qword ptr single_one
	movlps qword ptr tmpsingle, xmm0

ENDM

JJ      equ     10            ; lag 1
KK      equ     17            ; lag 2, size of circular buffer
R1      equ     19            ; rotate count
R2      equ     27            ; rotate count

.data

align   16
randp1  DT      1.5           ; used for conversion to float
        DW      0             ; alignment
p1      DD      0             ; pointer in circular buffer
p2      DD      0             ; pointer in circular buffer
randbuf DD      (2*KK) dup(?) ; circular buffer

tmpreal10	DT	0

align 16
tmpdouble 	REAL8 1.0, 1.0
double_one	REAL8	1.0, 1.0

align 16
tmpsingle		REAL4	1.0, 1.0
single_one	REAL4	1.0, 1.0

; Lookup table for M_WRandomL - uses 544 bytes
align 16
PtrNext label dword
	dd 8 dup ((KK-1)*8)
	counter = 0
  	REPEAT ((KK-1)*8)-1
    	dd counter
			counter = counter + 1
   	endm

; Used by M_WRandom2
align 4
	last_value	dd 0

.code
; 
start:
; 

	rdtsc
	push eax
	call WRandomInit	; Init to something
	
	; Print some samples of numbers
	M_WBRandom
	push edx
	print hex$(eax)
	pop edx
	print hex$(edx)
	print " (WBRandom)",13,10
	
	M_WBRandom
	push edx
	print hex$(eax)
	pop edx
	print hex$(edx)
	print " (WBRandom)",13,10
	
	print " ",13,10
	
	M_WBRandomL
	push edx
	print hex$(eax)
	pop edx
	print hex$(edx)
	print " (WBRandom - lookup)",13,10
	
	M_WBRandomL
	push edx
	print hex$(eax)
	pop edx
	print hex$(edx)
	print " (WBRandom - lookup)",13,10
	
	print " ",13,10
	
	M_WBRandom2
	print hex$(eax)
	print " (WBRandom - 32 bit)",13,10
	
	M_WBRandom2
	print hex$(eax)
	print " (WBRandom - 32 bit)",13,10
	
	print " ",13,10
	
	M_WBRandom2L
	print hex$(eax)
	print " (WBRandom - 32 bit & lookup)",13,10
	
	M_WBRandom2L
	print hex$(eax)
	print " (WBRandom - 32 bit & lookup)",13,10
	
	print " ",13,10
	
	M_WRandom
	fstp tmpreal10
	print real10$(tmpreal10)
	print " (WRandom, real10)",13,10
	
	M_WRandom
	fstp tmpreal10
	print real10$(tmpreal10)
	print " (WRandom, real10)",13,10
	
	print " ",13,10
	
	M_WIRandom 0, 1000
	print ustr$(eax)
	print " (WIRandom, 0-1000)",13,10
	
	M_WIRandom 0, 1000
	print ustr$(eax)
	print " (WIRandom, 0-1000)",13,10
	
	print " ",13,10
	
	M_WDRandom
	print real8$(tmpdouble)
	print " (WDRandom, double)",13,10
	
	M_WDRandom
	print real8$(tmpdouble)
	print " (WDRandom, double)",13,10
	
	print " ",13,10
	
	M_WSRandom
	print real4$(tmpsingle)
	print " (WSRandom, single)",13,10
	
	M_WSRandom
	print real4$(tmpsingle)
	print " (WSRandom, single)",13,10
	
	print " ",13,10
	
	counter_begin 10000000, HIGH_PRIORITY_CLASS
	
	M_WBRandom
	
	counter_end
	print ustr$(eax)
	print " cycles, WBRandom",13,10
	
	counter_begin 10000000, HIGH_PRIORITY_CLASS
	
	M_WBRandomL
	
	counter_end
	print ustr$(eax)
	print " cycles, WBRandom - lookup",13,10
	
	counter_begin 10000000, HIGH_PRIORITY_CLASS
	
	M_WBRandom2
	
	counter_end
	print ustr$(eax)
	print " cycles, WBRandom - 32 bit",13,10
	
	counter_begin 10000000, HIGH_PRIORITY_CLASS
	
	M_WBRandom2L
	
	counter_end
	print ustr$(eax)
	print " cycles, WBRandom - 32 bit & lookup",13,10
	
	counter_begin 10000000, HIGH_PRIORITY_CLASS
	
	M_WRandom
  
	counter_end
	print ustr$(eax)
	print " cycles, WRandom",13,10
	
	counter_begin 10000000, HIGH_PRIORITY_CLASS
	
	M_WIRandom 100, 100000
	
	counter_end
	print ustr$(eax)
	print " cycles, WIRandom",13,10
	
	counter_begin 10000000, HIGH_PRIORITY_CLASS
	
	M_WDRandom
	
	counter_end
	print ustr$(eax)
	print " cycles, WDRandom",13,10
	
	counter_begin 10000000, HIGH_PRIORITY_CLASS
	
	M_WSRandom
	
	counter_end
	print ustr$(eax)
	print " cycles, WSRandom",13,10
	
	inkey "Press any key to exit..."
	exit

; 

WRandomInit PROC seed:DWORD

        MOV     EAX, seed    ; seed
        XOR     ECX, ECX
        ; make random numbers and put them into buffer
R80:    IMUL    EAX, 2891336453
        INC     EAX
        MOV     [randbuf][ECX*4], EAX
        INC     ECX
        CMP     ECX, KK*2
        JB      R80
        FLD1
        FSTP    [randp1]        ; initialize bit 64-79
        MOV     [p1], 0         ; initialize buffer pointers
        MOV     [p2], JJ*8
        CALL    WBRandom           ; prepare first random number
        PUSH    EDI
        MOV     EDI, 30
R90:    CALL    WBRandom
        DEC     EDI
        JNZ     R90
        POP     EDI                                                

        RET
WRandomInit ENDP

WBRandom PROC

        PUSH    EBX
        MOV     EBX, [p1]       ; ring buffer pointers
        MOV     ECX, [p2]       ; ring buffer pointer
        MOV     EDX, [randbuf][EBX]
        MOV     EAX, [randbuf][EBX+4]       
R20:    ROL     EDX, R1                  ; rotate bits
        ROL     EAX, R2
        ADD     EDX, [randbuf][ECX]      ; add two dwords
        ADD     EAX, [randbuf][ECX+4]
        MOV     [randbuf][EBX], EAX      ; save in swapped order
        MOV     [randbuf][EBX+4], EDX
        SUB     EBX, 8                   ; decrement p1
        JNC     SHORT R30
        MOV     EBX, (KK-1)*8            ; wrap around p1
R30:    SUB     ECX, 8                   ; decrement p2
        JNC     SHORT R40
        MOV     ECX, (KK-1)*8            ; wrap around p2
R40:    MOV     [p1], EBX                ; save updated pointers
        MOV     [p2], ECX
        POP     EBX
        RET
        
WBRandom ENDP

end start
